package ru.dbotthepony.kstarbound.server.world

import it.unimi.dsi.fastutil.longs.Long2ObjectFunction
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectAVLTreeSet
import it.unimi.dsi.fastutil.objects.ReferenceLinkedOpenHashSet
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet
import ru.dbotthepony.kommons.collect.chainOptionalFutures
import ru.dbotthepony.kommons.core.KOptional
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.client.network.packets.JoinWorldPacket
import ru.dbotthepony.kstarbound.server.StarboundServer
import ru.dbotthepony.kstarbound.server.network.ServerConnection
import ru.dbotthepony.kstarbound.util.ExecutionSpinner
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.ICellChangeListener
import ru.dbotthepony.kstarbound.world.IChunkListener
import ru.dbotthepony.kstarbound.world.IEntityAdditionListener
import ru.dbotthepony.kstarbound.world.IEntityRemovalListener
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.WorldGeometry
import ru.dbotthepony.kstarbound.world.api.ImmutableCell
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import java.util.Collections
import java.util.concurrent.CompletableFuture
import java.util.concurrent.RejectedExecutionException
import java.util.concurrent.atomic.AtomicInteger
import java.util.concurrent.locks.LockSupport
import java.util.concurrent.locks.ReentrantLock
import java.util.function.Consumer
import java.util.function.Supplier
import kotlin.concurrent.withLock

class ServerWorld(
	val server: StarboundServer,
	seed: Long,
	geometry: WorldGeometry,
) : World<ServerWorld, ServerChunk>(seed, geometry) {
	init {
		server.worlds.add(this)
	}

	private val internalPlayers = ArrayList<ServerConnection>()
	val players: List<ServerConnection> = Collections.unmodifiableList(internalPlayers)

	private fun doAcceptPlayer(player: ServerConnection): Boolean {
		if (player !in internalPlayers) {
			internalPlayers.add(player)
			player.world?.removePlayer(player)
			player.world = this
			player.send(JoinWorldPacket(this))
			return true
		}

		return false
	}

	fun acceptPlayer(player: ServerConnection): CompletableFuture<Boolean> {
		try {
			return CompletableFuture.supplyAsync(Supplier { doAcceptPlayer(player) }, mailbox)
		} catch (err: RejectedExecutionException) {
			return CompletableFuture.completedFuture(false)
		}
	}

	private fun doRemovePlayer(player: ServerConnection): Boolean {
		return internalPlayers.remove(player)
	}

	fun removePlayer(player: ServerConnection): CompletableFuture<Boolean> {
		try {
			return CompletableFuture.supplyAsync(Supplier { doRemovePlayer(player) }, mailbox)
		} catch (err: RejectedExecutionException) {
			return CompletableFuture.completedFuture(false)
		}
	}

	val spinner = ExecutionSpinner(mailbox, ::spin, Starbound.TICK_TIME_ADVANCE_NANOS)
	val thread = Thread(spinner, "Starbound Server World $seed")
	val ticketListLock = ReentrantLock()

	@Volatile
	var isClosed: Boolean = false
		private set

	init {
		thread.isDaemon = true
	}

	private val chunkProviders = ArrayList<IChunkSource>()
	var saver: IChunkSaver? = null

	fun addChunkSource(source: IChunkSource) {
		chunkProviders.add(source)
	}

	override fun close() {
		if (!isClosed) {
			super.close()
			isClosed = true

			lock.withLock {
				internalPlayers.forEach {
					it.world = null
				}
			}

			LockSupport.unpark(thread)
		}
	}

	private fun spin(): Boolean {
		if (isClosed) return false

		try {
			think()
			return true
		} catch (err: Throwable) {
			close()
			return false
		}
	}

	override val isRemote: Boolean
		get() = false

	override fun isSameThread(): Boolean {
		return Thread.currentThread() === thread
	}

	override fun thinkInner() {
		internalPlayers.forEach { if (!isClosed) it.tick() }

		ticketListLock.withLock {
			ticketLists.removeIf {
				val valid = it.tick()

				if (!valid) {
					val removed = ticketMap.remove(it.pos.toLong())
					check(removed == it) { "Expected to remove $it, but removed $removed" }

					val chunk = chunkMap[it.pos]

					if (chunk != null) {
						val unloadable = chunk.entities.filter { it.isApplicableForUnloading }

						saver?.saveCells(it.pos, chunk.copyCells())
						saver?.saveEntities(it.pos, unloadable)

						unloadable.forEach {
							it.remove()
						}

						chunkMap.remove(it.pos)
					}
				}

				!valid
			}
		}
	}

	override fun chunkFactory(pos: ChunkPos): ServerChunk {
		return ServerChunk(this, pos)
	}

	private val ticketMap = Long2ObjectOpenHashMap<TicketList>()
	private val ticketLists = ArrayList<TicketList>()

	private fun getTicketList(pos: ChunkPos): TicketList {
		return ticketMap.computeIfAbsent(geometry.wrapToLong(pos), Long2ObjectFunction { TicketList(it) })
	}

	fun permanentChunkTicket(pos: ChunkPos): ITicket {
		lock.withLock {
			return getTicketList(pos).Ticket()
		}
	}

	fun temporaryChunkTicket(pos: ChunkPos, time: Int): ITicket {
		require(time > 0) { "Invalid ticket time: $time" }

		lock.withLock {
			return getTicketList(pos).TimedTicket(time)
		}
	}

	interface ITicket {
		fun cancel()
		val isCanceled: Boolean
		val pos: ChunkPos
		val id: Int

		val chunk: ServerChunk?

		fun addListener(listener: IChunkListener)
		fun removeListener(listener: IChunkListener)
	}

	interface ITimedTicket : ITicket, Comparable<ITimedTicket> {
		val timeRemaining: Int
		fun prolong(ticks: Int)

		override fun compareTo(other: ITimedTicket): Int {
			val cmp = timeRemaining.compareTo(other.timeRemaining)
			if (cmp != 0) return cmp
			return id.compareTo(other.id)
		}
	}

	private inner class TicketList(val pos: ChunkPos) : IChunkListener, IChunkMapListener<ServerChunk> {
		constructor(pos: Long) : this(ChunkPos(pos))

		private var first = true
		private val permanent = ArrayList<Ticket>()
		private val temporary = ObjectAVLTreeSet<TimedTicket>()
		private var ticks = 0
		private var nextTicketID = AtomicInteger()
		private var weAreResponsibleForLoadingTheChunk = false

		val isValid: Boolean
			get() = temporary.isNotEmpty() || permanent.isNotEmpty()

		fun tick(): Boolean {
			ticks++

			while (temporary.isNotEmpty() && temporary.first().timeRemaining <= 0) {
				val ticket = temporary.first()
				ticket.isCanceled = true
				temporary.remove(ticket)
			}

			return temporary.isNotEmpty() || permanent.isNotEmpty()
		}

		override fun onEntityAdded(entity: AbstractEntity) {
			permanent.forEach { it.onEntityAdded(entity) }
			temporary.forEach { it.onEntityAdded(entity) }
		}

		override fun onEntityRemoved(entity: AbstractEntity) {
			permanent.forEach { it.onEntityRemoved(entity) }
			temporary.forEach { it.onEntityRemoved(entity) }
		}

		override fun onCellChanges(x: Int, y: Int, cell: ImmutableCell) {
			permanent.forEach { it.onCellChanges(x, y, cell) }
			temporary.forEach { it.onCellChanges(x, y, cell) }
		}

		override fun onChunkCreated(chunk: ServerChunk) {
			if (chunk.pos == pos) {
				chunk.addListener(this)
			}
		}

		override fun onChunkRemoved(chunk: ServerChunk) {
			if (chunk.pos == pos) {
				chunk.removeListener(this)
			}
		}

		abstract inner class AbstractTicket : ITicket, IChunkListener {
			final override val id: Int = nextTicketID.getAndIncrement()
			final override val pos: ChunkPos
				get() = this@TicketList.pos

			final override var isCanceled: Boolean = false

			fun init() {
				if (first) {
					first = false

					if (geometry.x.inBoundsChunk(pos.x) && geometry.y.inBoundsChunk(pos.y)) {
						ticketLists.add(this@TicketList)
						chunkMap.addListener(this@TicketList)

						if (chunkProviders.isNotEmpty() && chunkMap[pos] == null) {
							weAreResponsibleForLoadingTheChunk = true

							chainOptionalFutures(chunkProviders)
							{ if (!isValid) CompletableFuture.completedFuture(KOptional.empty()) else it.getTiles(pos) }
								.thenAccept(Consumer { tiles ->
									if (!isValid || !tiles.isPresent) return@Consumer

									chainOptionalFutures(chunkProviders)
									{ if (!isValid) CompletableFuture.completedFuture(KOptional.empty()) else it.getEntities(pos) }
										.thenAcceptAsync(Consumer { ents ->
											if (!isValid) return@Consumer
											val chunk = chunkMap.compute(pos) ?: return@Consumer
											chunk.loadCells(tiles.value)

											ents.ifPresent {
												for (obj in it) {
													obj.spawn(this@ServerWorld)
												}
											}
										}, mailbox)
								})
						}
					}
				}
			}

			final override fun cancel() {
				if (isCanceled) return

				lock.withLock {
					if (isCanceled) return
					isCanceled = true
					chunk?.entities?.forEach { e -> listeners.forEach { it.onEntityRemoved(e) } }
					onCancel()
				}
			}

			protected abstract fun onCancel()
			final override val chunk: ServerChunk?
				get() = chunkMap[pos]

			private val listeners = ReferenceLinkedOpenHashSet<IChunkListener>()

			final override fun addListener(listener: IChunkListener) {
				if (isCanceled) return
				listeners.add(listener)
				chunk?.entities?.forEach { listener.onEntityAdded(it) }
			}

			final override fun removeListener(listener: IChunkListener) {
				if (listeners.remove(listener)) {
					chunk?.entities?.forEach { listener.onEntityRemoved(it) }
				}
			}

			final override fun onEntityAdded(entity: AbstractEntity) {
				listeners.forEach { it.onEntityAdded(entity) }
			}

			final override fun onEntityRemoved(entity: AbstractEntity) {
				listeners.forEach { it.onEntityRemoved(entity) }
			}

			final override fun onCellChanges(x: Int, y: Int, cell: ImmutableCell) {
				listeners.forEach { it.onCellChanges(x, y, cell) }
			}
		}

		inner class Ticket : AbstractTicket() {
			init {
				permanent.add(this)
				init()
			}

			override fun onCancel() {
				permanent.remove(this)
			}
		}

		inner class TimedTicket(expiresAt: Int) : AbstractTicket(), ITimedTicket {
			var expiresAt = expiresAt + ticks

			override val timeRemaining: Int
				get() = (expiresAt - ticks).coerceAtLeast(0)

			init {
				temporary.add(this)
				init()
			}

			override fun onCancel() {
				temporary.remove(this)
			}

			override fun prolong(ticks: Int) {
				if (ticks == 0 || isCanceled) return

				lock.withLock {
					if (isCanceled) return

					temporary.remove(this)
					expiresAt += ticks
					if (timeRemaining > 0) temporary.add(this)
				}
			}
		}
	}
}
